using System;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;
using NUnit.Framework.Constraints;

namespace Smdn.Net.Connection {
  [TestFixture]
  public class ConnectionBaseTests {
    private class PseudoConnection : ConnectionBase {
      protected override TraceSource TraceSource => null;

      private readonly Func<bool, int> getDefaultPort;

      public PseudoConnection() { }

      public PseudoConnection(Func<bool, int> getDefaultPort)
      {
        this.getDefaultPort = getDefaultPort;
      }

      protected override int GetDefaultPort(bool asSecurePortConnection)
        => getDefaultPort == null ? base.GetDefaultPort(asSecurePortConnection) : getDefaultPort(asSecurePortConnection);

      public Stream ConnectionStream => base.Stream;

      public void Connect(IPEndPoint endPoint)
      {
        base.Connect(endPoint.Address.ToString(), endPoint.Port, null, null);
      }

      public Task ConnectAsync(IPEndPoint endPoint)
      {
        return base.ConnectAsync(endPoint.Address.ToString(), endPoint.Port, null, null);
      }

      public void ConnectAndUpgrade(IPEndPoint endPoint, UpgradeConnectionStreamCallback upgradeConnectionStream)
      {
        base.Connect(endPoint.Address.ToString(), endPoint.Port, null, upgradeConnectionStream);
      }

      public Task ConnectAndUpgradeAsync(IPEndPoint endPoint, UpgradeConnectionStreamCallback upgradeConnectionStream)
      {
        return base.ConnectAsync(endPoint.Address.ToString(), endPoint.Port, null, upgradeConnectionStream);
      }
    }

    private class PseudoServer : IDisposable {
      private TcpListener server;
      private Task acceptClientTask;
      private CancellationTokenSource cts;

      public IPEndPoint EndPoint => server?.LocalEndpoint as IPEndPoint;
      public bool IsClientConnected => acceptClientTask != null && !acceptClientTask.Wait(100);

      public PseudoServer()
      {
        server = new TcpListener(new IPEndPoint(IPAddress.Loopback, 0));
      }

      public void Start()
      {
        server.Start();
        cts = new CancellationTokenSource();

        acceptClientTask = Task.Run(() => AcceptClient(cts.Token));
      }

      public void StartNotAccepting()
      {
        server.Start();
      }

      private static readonly byte[] peekClientBuffer = new byte[1];

      private void AcceptClient(CancellationToken cancellationToken)
      {
        var t = server.AcceptTcpClientAsync();

        t.Wait(cancellationToken);

        var client = t.Result;

        for (; ; ) {
          if (cancellationToken.IsCancellationRequested)
            break;

          if (!client.Connected)
            break;

          if (client.Client.Poll(0, SelectMode.SelectRead))
            if (client.Client.Receive(peekClientBuffer, SocketFlags.Peek) == 0)
              break; // client disconnected

          Thread.Sleep(10);
        }
      }

      public void Dispose()
      {
        cts?.Cancel();
        cts?.Dispose();
        cts = null;

        try {
          acceptClientTask?.Wait();
          acceptClientTask = null;
        }
        catch (AggregateException) {
          // ignore
        }

        server?.Stop();
        server = null;
      }
    }

    // https://stackoverflow.com/questions/39327013/nunit-assertion-that-a-certain-exception-is-not-thrown
    private class ThrowsSocketExceptionConstraint : ThrowsExceptionConstraint {
#if NETCOREAPP2_0 || NETCOREAPP2_1
      public override string Description => "throws SocketException or SocketExceptionFactory+ExtendedSocketException";
#else
      public override string Description => "throws SocketException";
#endif

      public override ConstraintResult ApplyTo<TActual>(TActual actual)
      {
        var result = base.ApplyTo<TActual>(actual);

        return new ThrowsSocketExceptionConstraintResult(this, result.ActualValue as Exception);
      }

      protected override object GetTestObject<TActual>(ActualValueDelegate<TActual> del)
      {
        return new TestDelegate(() => del());
      }

      class ThrowsSocketExceptionConstraintResult : ConstraintResult {
        private readonly Exception caughtException;

        public ThrowsSocketExceptionConstraintResult(ThrowsSocketExceptionConstraint constraint, Exception caughtException)
          : base(constraint,
                 caughtException,
#if NETCOREAPP2_0 || NETCOREAPP2_1
                 caughtException is SocketException || caughtException.GetType().FullName == "System.Net.Internals.SocketExceptionFactory+ExtendedSocketException"
#else
                 caughtException is SocketException
#endif
                 )
        {
          this.caughtException = caughtException;
        }

        public override void WriteActualValueTo(MessageWriter writer)
        {
          if (this.Status == ConstraintStatus.Failure)
            writer.Write("thrown {0}", caughtException.GetType().FullName);
          else
            base.WriteActualValueTo(writer);
        }
      }
    }

#if NET45
    private static Task<T> CreateTaskFromException<T>(Exception exception)
    {
      var tcs = new TaskCompletionSource<T>();

      tcs.SetException(exception);

      return tcs.Task;
    }
#endif

    [Test]
    public void TestConstruct()
    {
      using (var connection = new PseudoConnection()) {
        TestPropertyNotConnected(connection);

        Assert.DoesNotThrow(() => connection.ReceiveTimeout = 42, nameof(connection.ReceiveTimeout));
        Assert.DoesNotThrow(() => connection.SendTimeout = 42, nameof(connection.SendTimeout));

        Assert.DoesNotThrow(() => Assert.AreEqual(connection.ReceiveTimeout, 42, nameof(connection.ReceiveTimeout)));
        Assert.DoesNotThrow(() => Assert.AreEqual(connection.SendTimeout, 42, nameof(connection.SendTimeout)));
      }
    }

    private void TestPropertyNotConnected(ConnectionBase connection)
    {
      Assert.IsNotNull(connection.Id, nameof(connection.Id));
      Assert.DoesNotThrow(() => Assert.IsFalse(connection.IsConnected), nameof(connection.IsConnected));

      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(connection.Host), nameof(connection.Host));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(connection.Port), nameof(connection.Port));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(connection.LocalEndPoint), nameof(connection.LocalEndPoint));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(connection.RemoteEndPoint), nameof(connection.RemoteEndPoint));
      Assert.Throws<InvalidOperationException>(() => Assert.IsFalse(connection.IsSecureConnection), nameof(connection.IsSecureConnection));
      Assert.Throws<InvalidOperationException>(() => Assert.IsFalse(connection.IsSecurePortConnection), nameof(connection.IsSecurePortConnection));
    }

    [Test]
    public void TestSendTimeoutReceiveTimeout_OutOfRange_NotConnected()
    {
      using (var connection = new PseudoConnection()) {
        Assert.AreEqual(nameof(connection.ReceiveTimeout),
                        Assert.Throws<ArgumentOutOfRangeException>(() => connection.ReceiveTimeout = -2).ParamName);
        Assert.AreEqual(nameof(connection.ReceiveTimeout),
                        Assert.Throws<ArgumentOutOfRangeException>(() => connection.ReceiveTimeout = int.MinValue).ParamName);
        Assert.AreEqual(nameof(connection.SendTimeout),
                        Assert.Throws<ArgumentOutOfRangeException>(() => connection.SendTimeout = -2).ParamName);
        Assert.AreEqual(nameof(connection.SendTimeout),
                        Assert.Throws<ArgumentOutOfRangeException>(() => connection.SendTimeout = int.MinValue).ParamName);
      }
    }

    [Test]
    public void TestSendTimeoutReceiveTimeout_OutOfRange_Connected()
    {
      using (var server = new PseudoServer()) {
        server.Start();
        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);

          Assert.AreEqual(nameof(connection.ReceiveTimeout),
                          Assert.Throws<ArgumentOutOfRangeException>(() => connection.ReceiveTimeout = -2).ParamName);
          Assert.AreEqual(nameof(connection.ReceiveTimeout),
                          Assert.Throws<ArgumentOutOfRangeException>(() => connection.ReceiveTimeout = int.MinValue).ParamName);
          Assert.AreEqual(nameof(connection.SendTimeout),
                          Assert.Throws<ArgumentOutOfRangeException>(() => connection.SendTimeout = -2).ParamName);
          Assert.AreEqual(nameof(connection.SendTimeout),
                          Assert.Throws<ArgumentOutOfRangeException>(() => connection.SendTimeout = int.MinValue).ParamName);
        }
      }
    }

    [Test]
    public void TestSendTimeoutReceiveTimeout_Valid_NotConnected()
    {
      using (var connection = new PseudoConnection()) {
        Assert.DoesNotThrow(() => connection.ReceiveTimeout = -1);
        Assert.AreEqual(-1, connection.ReceiveTimeout);
        Assert.DoesNotThrow(() => connection.ReceiveTimeout = 0);
        Assert.AreEqual(0, connection.ReceiveTimeout);
        Assert.DoesNotThrow(() => connection.ReceiveTimeout = 1);
        Assert.AreEqual(1, connection.ReceiveTimeout);
        Assert.DoesNotThrow(() => connection.ReceiveTimeout = int.MaxValue);
        Assert.AreEqual(int.MaxValue, connection.ReceiveTimeout);

        Assert.DoesNotThrow(() => connection.SendTimeout = -1);
        Assert.AreEqual(-1, connection.SendTimeout);
        Assert.DoesNotThrow(() => connection.SendTimeout = 0);
        Assert.AreEqual(0, connection.SendTimeout);
        Assert.DoesNotThrow(() => connection.SendTimeout = 1);
        Assert.AreEqual(1, connection.SendTimeout);
        Assert.DoesNotThrow(() => connection.SendTimeout = int.MaxValue);
        Assert.AreEqual(int.MaxValue, connection.SendTimeout);
      }
    }

    [Test]
    public void TestSendTimeoutReceiveTimeout_Valid_Connected()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);

          Assert.DoesNotThrow(() => connection.ReceiveTimeout = -1);
          Assert.AreEqual(-1, connection.ReceiveTimeout);
          Assert.DoesNotThrow(() => connection.ReceiveTimeout = 0);
          Assert.AreEqual(0, connection.ReceiveTimeout);
          Assert.DoesNotThrow(() => connection.ReceiveTimeout = 1);
          Assert.AreEqual(1, connection.ReceiveTimeout);
          Assert.DoesNotThrow(() => connection.ReceiveTimeout = int.MaxValue);
          Assert.AreEqual(int.MaxValue, connection.ReceiveTimeout);

          Assert.DoesNotThrow(() => connection.SendTimeout = -1);
          Assert.AreEqual(-1, connection.SendTimeout);
          Assert.DoesNotThrow(() => connection.SendTimeout = 0);
          Assert.AreEqual(0, connection.SendTimeout);
          Assert.DoesNotThrow(() => connection.SendTimeout = 1);
          Assert.AreEqual(1, connection.SendTimeout);
          Assert.DoesNotThrow(() => connection.SendTimeout = int.MaxValue);
          Assert.AreEqual(int.MaxValue, connection.SendTimeout);
        }
      }
    }

    [Test]
    public void TestSendTimeoutReceiveTimeout_SetBeforeConnect()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.ReceiveTimeout = 1;
          connection.SendTimeout = 2;

          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);

          Assert.AreEqual(1, connection.ReceiveTimeout);
          Assert.AreEqual(2, connection.SendTimeout);

          connection.Close();

          Assert.AreEqual(1, connection.ReceiveTimeout);
          Assert.AreEqual(2, connection.SendTimeout);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestSendTimeoutReceiveTimeout_SetAfterConnect()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);

          connection.ReceiveTimeout = 1;
          connection.SendTimeout = 2;

          Assert.AreEqual(1, connection.ReceiveTimeout);
          Assert.AreEqual(2, connection.SendTimeout);

          connection.Close();

          Assert.AreEqual(1, connection.ReceiveTimeout);
          Assert.AreEqual(2, connection.SendTimeout);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestConnect() => Connect((conn, endpoint) => Assert.DoesNotThrow(() => conn.Connect(endpoint)));

    [Test]
    public void TestConnectAsync() => Connect((conn, endpoint) => Assert.DoesNotThrowAsync(async () => await conn.ConnectAsync(endpoint)));

    private void Connect(Action<PseudoConnection, IPEndPoint> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          testAction(connection, server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecureConnection);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.AreEqual(server.EndPoint, connection.RemoteEndPoint);
          Assert.IsNotNull(connection.LocalEndPoint);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestConnect_AlreadyConnected()
      => Connect_AlreadyConnected(conn => conn.Connect("host.test", 0));

    [Test]
    public void TestConnectAsync_AlreadyConnected()
      => Connect_AlreadyConnected(conn => conn.ConnectAsync("host.test", 0));

    private void Connect_AlreadyConnected(Action<PseudoConnection> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.AreEqual(server.EndPoint, connection.RemoteEndPoint);

          Assert.Throws<InvalidOperationException>(() => testAction(connection));

          Assert.IsTrue(connection.IsConnected);
          Assert.AreEqual(server.EndPoint, connection.RemoteEndPoint);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestConnect_ReconnectAfterClosed()
      => Connect_ReconnectAfterClosed((conn, endpoint) => Assert.DoesNotThrow(() => conn.Connect(endpoint)));

    [Test]
    public void TestConnectAsync_ReconnectAfterClosed()
      => Connect_ReconnectAfterClosed((conn, endpoint) => Assert.DoesNotThrowAsync(async () => await conn.ConnectAsync(endpoint)));

    private void Connect_ReconnectAfterClosed(Action<PseudoConnection, IPEndPoint> testAction)
    {
      using (var connection = new PseudoConnection()) {
        using (var server = new PseudoServer()) {
          server.Start();

          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);

          connection.Close();

          Assert.IsFalse(connection.IsConnected);
        }

        // connect again
        using (var server = new PseudoServer()) {
          server.Start();

          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
        }
      }
    }

    [Test]
    public void TestConnect_GetDefaultPort()
      => Connect_GetDefaultPort((conn, host) => Assert.DoesNotThrow(() => conn.Connect(host, -1)));

    [Test]
    public void TestConnectAsync_GetDefaultPort()
      => Connect_GetDefaultPort((conn, host) => Assert.DoesNotThrowAsync(async () => await conn.ConnectAsync(host, -1)));

    private void Connect_GetDefaultPort(Action<PseudoConnection, string> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        var getDefaultPort = new Func<bool, int>(asSecurePortConnection => {
          Assert.IsFalse(asSecurePortConnection, nameof(asSecurePortConnection));
          return server.EndPoint.Port;
        });

        using (var connection = new PseudoConnection(getDefaultPort)) {
          testAction(connection, server.EndPoint.Address.ToString());

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);
          Assert.AreEqual(server.EndPoint, connection.RemoteEndPoint);
        }
      }
    }

    [Test]
    public void TestConnect_GetDefaultPort_Undefined()
      => Connect_GetDefaultPort_Undefined((conn, host) => conn.Connect(host, -1));

    [Test]
    public void TestConnectAsync_GetDefaultPort_Undefined()
      => Connect_GetDefaultPort_Undefined((conn, host) => conn.ConnectAsync(host, -1));

    private void Connect_GetDefaultPort_Undefined(Action<PseudoConnection, string> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection(getDefaultPort: null)) {
          var upgradeConnection = new UpgradeConnectionStreamCallback((c, s, ct) => {
            Assert.AreSame(c, connection);
            Assert.AreSame(s, connection.ConnectionStream);

            return Task.FromResult<Stream>(new PseudoSslStream(s));
          });

          var ex = Assert.Throws<ArgumentOutOfRangeException>(() => testAction(connection, server.EndPoint.Address.ToString()));

          Assert.AreEqual("port", ex.ParamName);
          Assert.AreEqual(-1, ex.ActualValue);

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);
        }
      }
    }

    [Test]
    public void TestConnect_GetDefaultPort_AsSecurePortConnection()
      => Connect_GetDefaultPort_AsSecurePortConnection((conn, host, upgradeConnection) => Assert.DoesNotThrow(() => conn.Connect(host, -1, null, upgradeConnection)));

    [Test]
    public void TestConnectAsync_GetDefaultPort_AsSecurePortConnection()
      => Connect_GetDefaultPort_AsSecurePortConnection((conn, host, upgradeConnection) => Assert.DoesNotThrowAsync(async () => await conn.ConnectAsync(host, -1, null, upgradeConnection)));

    private void Connect_GetDefaultPort_AsSecurePortConnection(Action<PseudoConnection, string, UpgradeConnectionStreamCallback> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        var getDefaultPort = new Func<bool, int>(asSecurePortConnection => {
          Assert.IsTrue(asSecurePortConnection, nameof(asSecurePortConnection));
          return server.EndPoint.Port;
        });

        using (var connection = new PseudoConnection(getDefaultPort)) {
          var upgradeConnection = new UpgradeConnectionStreamCallback((c, s, ct) => {
            Assert.AreSame(c, connection);
            Assert.AreSame(s, connection.ConnectionStream);

            return Task.FromResult<Stream>(new PseudoSslStream(s));
          });

          testAction(connection, server.EndPoint.Address.ToString(), upgradeConnection);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsTrue(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.AreEqual(server.EndPoint, connection.RemoteEndPoint);
        }
      }
    }

    [Test]
    public void TestConnect_SocketException_ConnectionRefused()
      => Connect_SocketException_ConnectionRefused((conn, endpoint) => Assert.That(() => conn.Connect(endpoint), new ThrowsSocketExceptionConstraint()));

    [Test]
    public void TestConnectAsync_SocketException_ConnectionRefused()
      => Connect_SocketException_ConnectionRefused((conn, endpoint) => Assert.That(async () => await conn.ConnectAsync(endpoint), new ThrowsSocketExceptionConstraint()));

    private void Connect_SocketException_ConnectionRefused(Action<PseudoConnection, IPEndPoint> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        var endPoint = server.EndPoint;

        server.Dispose();

        using (var connection = new PseudoConnection()) {
          testAction(connection, endPoint);

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);
        }
      }
    }

    [Test]
    public void TestConnect_SocketException_NameResolution()
      => Connect_SocketException_NameResolution((conn) => Assert.That(() => conn.Connect("host.invalid", 1), new ThrowsSocketExceptionConstraint()));

    [Test]
    public void TestConnectAsync_SocketException_NameResolution()
      => Connect_SocketException_NameResolution((conn) => Assert.That(async () => await conn.ConnectAsync("host.invalid", 1), new ThrowsSocketExceptionConstraint()));

    private void Connect_SocketException_NameResolution(Action<PseudoConnection> testAction)
    {
      using (var connection = new PseudoConnection()) {
        testAction(connection);

        Assert.IsFalse(connection.IsConnected);

        TestPropertyNotConnected(connection);
      }
    }

    [Test]
    public void TestConnect_RequestCancellation_NameResolution()
    {
      Connect_RequestCancellation_NameResolution((conn, timeout) => {
        Assert.That(() => {
          using (var cts = new CancellationTokenSource(timeout)) {
            conn.Connect("nonexistent.test", 10000, null, null, cts.Token);
          }
        }, Throws.InstanceOf<OperationCanceledException>());
      });
    }

    [Test]
    public void TestConnectAsync_RequestCancellation_NameResolution()
    {
      Connect_RequestCancellation_NameResolution((conn, timeout) => {
        Assert.ThrowsAsync<OperationCanceledException>(async () => {
          using (var cts = new CancellationTokenSource(timeout)) {
            await conn.ConnectAsync("nonexistent.test", 10000, null, null, cts.Token);
          }
        });
      });
    }

    private void Connect_RequestCancellation_NameResolution(Action<PseudoConnection, TimeSpan> testAction)
    {
      var timeout = TimeSpan.FromMilliseconds(250);
      var timeoutAcceptable = TimeSpan.FromMilliseconds(500);

      using (var connection = new PseudoConnection()) {
        TestUtils.Assert.Elapses(timeout, timeoutAcceptable, () => testAction(connection, timeout));

        Assert.IsFalse(connection.IsConnected);

        TestPropertyNotConnected(connection);
      }
    }

    private class PseudoSslStream : Stream {
      public PseudoSslStream(Stream baseStream)
      {
      }

      public override bool CanRead => throw new NotImplementedException();
      public override bool CanSeek => throw new NotImplementedException();
      public override bool CanWrite => throw new NotImplementedException();
      public override long Length => throw new NotImplementedException();
      public override long Position { get => throw new NotImplementedException(); set => throw new NotImplementedException(); }
      public override void Flush() => throw new NotImplementedException();
      public override int Read(byte[] buffer, int offset, int count) => throw new NotImplementedException();
      public override long Seek(long offset, SeekOrigin origin) => throw new NotImplementedException();
      public override void SetLength(long value) => throw new NotImplementedException();
      public override void Write(byte[] buffer, int offset, int count) => throw new NotImplementedException();
    }

    [Test]
    public void TestConnect_UpgradeConnection()
      => Connect_UpgradeConnection((conn, endpoint, upgradeConnection) => Assert.DoesNotThrow(() => conn.ConnectAndUpgrade(endpoint, upgradeConnection)));

    [Test]
    public void TestConnectAsync_UpgradeConnection()
      => Connect_UpgradeConnection((conn, endpoint, upgradeConnection) => Assert.DoesNotThrowAsync(async () => await conn.ConnectAndUpgradeAsync(endpoint, upgradeConnection)));

    private void Connect_UpgradeConnection(Action<PseudoConnection, IPEndPoint, UpgradeConnectionStreamCallback> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          var upgradeConnection = new UpgradeConnectionStreamCallback((c, s, ct) => {
            Assert.AreSame(c, connection);
            Assert.AreSame(s, connection.ConnectionStream);

            return Task.FromResult<Stream>(new PseudoSslStream(s));
          });

          testAction(connection, server.EndPoint, upgradeConnection);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsTrue(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoSslStream>(connection.ConnectionStream);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestConnect_UpgradeConnection_Exception()
      => Connect_UpgradeConnection_Exception((conn, endpoint, upgradeConnection) => Assert.Throws<NotSupportedException>(() => conn.ConnectAndUpgrade(endpoint, upgradeConnection)));

    [Test]
    public void TestConnectAsync_UpgradeConnection_Exception()
      => Connect_UpgradeConnection_Exception((conn, endpoint, upgradeConnection) => Assert.ThrowsAsync<NotSupportedException>(async () => await conn.ConnectAndUpgradeAsync(endpoint, upgradeConnection)));

    private void Connect_UpgradeConnection_Exception(Action<PseudoConnection, IPEndPoint, UpgradeConnectionStreamCallback> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          var upgradeConnection = new UpgradeConnectionStreamCallback((c, s, ct) => throw new NotSupportedException());

          testAction(connection, server.EndPoint, upgradeConnection);

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestConnect_ResolveHostAddress()
      => Connect_ResolveHostAddress((conn, port, resolveHostAddress) => Assert.DoesNotThrow(() => conn.Connect("host.invalid", port, resolveHostAddress, null)));

    [Test]
    public void TestConnectAsync_ResolveHostAddress()
      => Connect_ResolveHostAddress((conn, port, resolveHostAddress) => Assert.DoesNotThrowAsync(async () => await conn.ConnectAsync("host.invalid", port, resolveHostAddress, null)));

    private void Connect_ResolveHostAddress(Action<PseudoConnection, int, ResolveHostAddressCallback> testAction)
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
            Assert.AreEqual("host.invalid", hostname);

            return Task.FromResult(server.EndPoint.Address);
          });

          testAction(connection, server.EndPoint.Port, resolveHostAddress);

          Assert.IsTrue(connection.IsConnected);
          Assert.AreEqual(server.EndPoint, connection.RemoteEndPoint);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestConnect_ResolveHostAddress_Null()
      => Connect_ResolveHostAddress_Null((conn, resolveHostAddress) => Assert.Throws<InvalidOperationException>(() => conn.Connect("host.invalid", 0, resolveHostAddress, null)));

    [Test]
    public void TestConnectAsync_ResolveHostAddress_Null()
      => Connect_ResolveHostAddress_Null((conn, resolveHostAddress) => Assert.ThrowsAsync<InvalidOperationException>(async () => await conn.ConnectAsync("host.invalid", 0, resolveHostAddress, null)));

    private void Connect_ResolveHostAddress_Null(Action<PseudoConnection, ResolveHostAddressCallback> testAction)
    {
      using (var connection = new PseudoConnection()) {
        var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
          Assert.AreEqual("host.invalid", hostname);

          return Task.FromResult<IPAddress>(null);
        });

        testAction(connection, resolveHostAddress);

        Assert.IsFalse(connection.IsConnected);

        TestPropertyNotConnected(connection);
      }
    }

    [Test]
    public void TestConnect_ResolveHostAddress_Exception()
      => Connect_ResolveHostAddress_Exception((conn, resolveHostAddress) => Assert.Throws<NotImplementedException>(() => conn.Connect("host.invalid", 0, resolveHostAddress, null)));

    [Test]
    public void TestConnectAsync_ResolveHostAddress_Exception()
      => Connect_ResolveHostAddress_Exception((conn, resolveHostAddress) => Assert.ThrowsAsync<NotImplementedException>(async () => await conn.ConnectAsync("host.invalid", 0, resolveHostAddress, null)));

    private void Connect_ResolveHostAddress_Exception(Action<PseudoConnection, ResolveHostAddressCallback> testAction)
    {
      using (var connection = new PseudoConnection()) {
        var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
          Assert.AreEqual("host.invalid", hostname);

#if NET45
          return CreateTaskFromException<IPAddress>(new NotImplementedException());
#else
          return Task.FromException<IPAddress>(new NotImplementedException());
#endif
        });

        testAction(connection, resolveHostAddress);

        Assert.IsFalse(connection.IsConnected);

        TestPropertyNotConnected(connection);
      }
    }

    [Test]
    public void TestClose()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);

          connection.Close();

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);

          Assert.DoesNotThrow(() => connection.ToString());
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeToSecureConnectionAsync()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          Assert.DoesNotThrowAsync(async () => {
            await connection.UpgradeToSecureConnectionAsync((c, s, ct) => {
              Assert.AreSame(c, connection);
              Assert.AreSame(s, connection.ConnectionStream);

              return Task.FromResult<Stream>(new PseudoSslStream(s));
            });
          });

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoSslStream>(connection.ConnectionStream);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeToSecureConnectionAsync_ArgumentNull()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          Assert.Throws<ArgumentNullException>(() => connection.UpgradeToSecureConnectionAsync(null));

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeToSecureConnectionAsync_Exception()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          Assert.ThrowsAsync<NotSupportedException>(async () => {
            await connection.UpgradeToSecureConnectionAsync((c, s, ct) => throw new NotSupportedException());
          });

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeToSecureConnectionAsync_TaskException()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          Assert.ThrowsAsync<NotSupportedException>(async () => {
            await connection.UpgradeToSecureConnectionAsync((c, s, ct) => {
#if NET45
              return CreateTaskFromException<Stream>(new NotSupportedException());
#else
              return Task.FromException<Stream>(new NotSupportedException());
#endif
            });
          });

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeToSecureConnectionAsync_SecurePortConnection()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.ConnectAndUpgrade(server.EndPoint, (c, s, ct) => {
            Assert.AreSame(c, connection);
            Assert.AreSame(s, connection.ConnectionStream);

            return Task.FromResult<Stream>(new PseudoSslStream(s));
          });

          Assert.IsTrue(connection.IsConnected);
          Assert.IsTrue(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoSslStream>(connection.ConnectionStream);

          Assert.Throws<InvalidOperationException>(() => connection.UpgradeToSecureConnectionAsync((c, s, ct) => {
            Assert.AreSame(c, connection);
            Assert.AreSame(s, connection.ConnectionStream);

            return Task.FromResult<Stream>(new PseudoSslStream(s));
          }));

          Assert.IsTrue(connection.IsConnected);
          Assert.IsTrue(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoSslStream>(connection.ConnectionStream);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeToSecureConnectionAsync_AlreadyUpgraded()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          // upgrade
          Assert.DoesNotThrow(() => connection.UpgradeToSecureConnectionAsync((c, s, ct) => {
            Assert.AreSame(c, connection);
            Assert.AreSame(s, connection.ConnectionStream);

            return Task.FromResult<Stream>(new PseudoSslStream(s));
          }));

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoSslStream>(connection.ConnectionStream);

          // attempt to upgrade again
          Assert.Throws<InvalidOperationException>(() => connection.UpgradeToSecureConnectionAsync((c, s, ct) => {
            Assert.Fail("upgradeStream called");
#if NET45
            return CreateTaskFromException<Stream>(new AssertionException("upgradeStream called"));
#else
            return Task.FromException<Stream>(new AssertionException("upgradeStream called"));
#endif
          }));

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoSslStream>(connection.ConnectionStream);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    private class PseudoWrapperStream : Stream {
      public PseudoWrapperStream(Stream baseStream)
      {
      }

      public override bool CanRead => throw new NotImplementedException();
      public override bool CanSeek => throw new NotImplementedException();
      public override bool CanWrite => throw new NotImplementedException();
      public override long Length => throw new NotImplementedException();
      public override long Position { get => throw new NotImplementedException(); set => throw new NotImplementedException(); }
      public override void Flush() => throw new NotImplementedException();
      public override int Read(byte[] buffer, int offset, int count) => throw new NotImplementedException();
      public override long Seek(long offset, SeekOrigin origin) => throw new NotImplementedException();
      public override void SetLength(long value) => throw new NotImplementedException();
      public override void Write(byte[] buffer, int offset, int count) => throw new NotImplementedException();
    }

    [Test]
    public void TestUpgradeConnectionAsync()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          Assert.DoesNotThrowAsync(async () => {
            await connection.UpgradeConnectionAsync((c, s, ct) => {
              Assert.AreSame(c, connection);
              Assert.AreSame(s, connection.ConnectionStream);

              return Task.FromResult<Stream>(new PseudoWrapperStream(s));
            });
          });

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoWrapperStream>(connection.ConnectionStream);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeConnectionAsync_ArgumentNull()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          Assert.Throws<ArgumentNullException>(() => connection.UpgradeConnectionAsync(null));

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeConnectionAsync_Exception()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          Assert.ThrowsAsync<NotSupportedException>(async () => {
            await connection.UpgradeConnectionAsync((c, s, ct) => throw new NotSupportedException());
          });

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeConnectionAsync_TaskException()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.Connect(server.EndPoint);

          Assert.IsTrue(connection.IsConnected);
          Assert.IsFalse(connection.IsSecurePortConnection);
          Assert.IsFalse(connection.IsSecureConnection);

          Assert.IsInstanceOf<NetworkStream>(connection.ConnectionStream);

          Assert.ThrowsAsync<NotSupportedException>(async () => {
            await connection.UpgradeConnectionAsync((c, s, ct) => {
#if NET45
              return CreateTaskFromException<Stream>(new NotSupportedException());
#else
              return Task.FromException<Stream>(new NotSupportedException());
#endif
            });
          });

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test]
    public void TestUpgradeConnectionAsync_SecurePortConnection()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          connection.ConnectAndUpgrade(server.EndPoint, (c, s, ct) => {
            Assert.AreSame(c, connection);
            Assert.AreSame(s, connection.ConnectionStream);

            return Task.FromResult<Stream>(new PseudoSslStream(s));
          });

          Assert.IsTrue(connection.IsConnected);
          Assert.IsTrue(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoSslStream>(connection.ConnectionStream);

          Assert.DoesNotThrowAsync(async () => {
            await connection.UpgradeConnectionAsync((c, s, ct) => {
              Assert.AreSame(c, connection);
              Assert.AreSame(s, connection.ConnectionStream);

              return Task.FromResult<Stream>(new PseudoWrapperStream(s));
            });
          });

          Assert.IsTrue(connection.IsConnected);
          Assert.IsTrue(connection.IsSecurePortConnection);
          Assert.IsTrue(connection.IsSecureConnection);

          Assert.IsInstanceOf<PseudoWrapperStream>(connection.ConnectionStream);
        }

        Assert.IsFalse(server.IsClientConnected);
      }
    }

    [Test, Ignore("can not test")]
    public void TestCreateClientSslStreamAsync_CancelOperation()
    {
      using (var server = new PseudoServer()) {
        server.Start();

        using (var connection = new PseudoConnection()) {
          var timeout = TimeSpan.FromMilliseconds(250);
          var timeoutAcceptable = TimeSpan.FromMilliseconds(500);

          connection.SendTimeout = 100;
          connection.ReceiveTimeout = 100;

          TestUtils.Assert.Elapses(timeout, timeoutAcceptable, () => {
            using (var cts = new CancellationTokenSource(timeout)) {
              Assert.ThrowsAsync<OperationCanceledException>(async () => {
                await connection.ConnectAsync(server.EndPoint.Address.ToString(), server.EndPoint.Port, null, (c, s, ct) => {
                  return ConnectionBase.CreateClientSslStreamAsync(c,
                                                                   s,
                                                                   new X509Certificate2Collection(),
                                                                   SslProtocols.None,
                                                                   (sender, certificate, chain, sslPolicyErrors) => true,
                                                                   null,
                                                                   ct);
                }, cts.Token);
              });
            }
          });

          Assert.IsFalse(connection.IsConnected);

          TestPropertyNotConnected(connection);
        }
      }
    }
  }
}
