using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Net.MessageAccessProtocols.Strings {
  [TestFixture]
  public class TypedStringSetTests {
    private class TypedString : IEquatable<TypedString> {
      public string Value { get; }

      public TypedString(string val)
      {
        Value = val;
      }

      public bool Equals(TypedString other) => TypedStringEqualityComparer.Comparer.Equals(this, other);
      public override string ToString() => Value;
    }

    private class TypedStringEqualityComparer : TypedStringEqualityComparer<TypedString> {
      protected override string GetStringExpressionOf(TypedString typedString) => typedString?.Value;
      public override bool Equals(string x, string y) => StringComparer.OrdinalIgnoreCase.Equals(x, y);
      public override int GetHashCode(string obj) => StringComparer.OrdinalIgnoreCase.GetHashCode(obj);

      public static readonly TypedStringEqualityComparer Comparer = new TypedStringEqualityComparer();
    }

    private class TestTypedStringSet : TypedStringSet<TypedString> {
      public TestTypedStringSet()
        : base(TypedStringEqualityComparer.Comparer)
      {
      }

      public TestTypedStringSet(bool readOnly, IEnumerable<TypedString> values)
        : base(values, TypedStringEqualityComparer.Comparer, asReadOnly: readOnly)
      {
      }
    }

    [Test]
    public void TestConstruct()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.IsFalse(set.IsReadOnly);
      Assert.AreEqual(3, set.Count);
    }

    [Test]
    public void TestConstructContainsSameKey()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("IMAP4rev1"),
      });

      Assert.AreEqual(1, set.Count);
      Assert.IsTrue(set.Contains("IMAP4rev1"));
    }

    [Test]
    public void TestContains()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.IsTrue(set.Contains(new TypedString("IMAP4rev1")));
      Assert.IsTrue(set.Contains("IMAP4rev1"));
      Assert.IsTrue(set.Contains(new TypedString("imap4REV1")));
      Assert.IsTrue(set.Contains("imap4REV1"));
      Assert.IsFalse(set.Contains(new TypedString("AUTH=DIGEST-MD5")));
      Assert.IsFalse(set.Contains("AUTH=DIGEST-MD5"));
    }

    [Test]
    public void TestContainsNull()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.Throws<ArgumentNullException>(() => set.Contains((TypedString)null));
      Assert.Throws<ArgumentNullException>(() => set.Contains((string)null));
    }

    [Test]
    public void TestTryGetValue()
    {
      var values = new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      };

      var set = new TestTypedStringSet(false, values);

      Assert.IsTrue(set.TryGetValue("imap4rev1", out var val0));
      Assert.AreSame(values[0], val0);

      Assert.IsTrue(set.TryGetValue(values[1].Value, out var val1));
      Assert.AreSame(values[1], val1);

      Assert.IsFalse(set.TryGetValue("AUTH=digest-md5", out var val2));
      Assert.IsNull(val2);
    }

    [Test]
    public void TestTryGetValueNull()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.Throws<ArgumentNullException>(() => set.TryGetValue(null, out _));
    }

    [Test]
    public void TestClear()
    {
      var values = new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      };

      var set = new TestTypedStringSet(false, values);

      Assert.IsTrue(set.TryGetValue("IMAP4rev1", out var _));
      Assert.AreEqual(3, set.Count);

      set.Clear();

      Assert.IsFalse(set.TryGetValue("IMAP4rev1", out var _));
      Assert.AreEqual(0, set.Count);
    }

    [Test]
    public void TestToArray()
    {
      var values = new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      };

      var set = new TestTypedStringSet(false, values);

      CollectionAssert.AreEquivalent(values, set.ToArray());
    }

    [Test]
    public void TestCopyTo()
    {
      var values = new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      };

      var set = new TestTypedStringSet(false, values);

      var copied = new TypedString[3];

      set.CopyTo(copied, 0);

      CollectionAssert.AreEquivalent(values, copied);
    }

    [Test]
    public void TestToString()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.AreEqual("IMAP4rev1, UIDPLUS, AUTH=PLAIN", set.ToString());
    }

    [Test]
    public void TestAdd()
    {
      var set = new TestTypedStringSet(false, Enumerable.Empty<TypedString>());

      Assert.IsFalse(set.Contains("IMAP4rev1"));
      Assert.AreEqual(0, set.Count);

      Assert.IsTrue(set.Add(new TypedString("IMAP4rev1")));

      Assert.AreEqual(1, set.Count);
      Assert.IsTrue(set.Contains("IMAP4rev1"));
    }

    [Test]
    public void TestAddNull()
    {
      var set = new TestTypedStringSet(false, Enumerable.Empty<TypedString>());

      Assert.Throws<ArgumentNullException>(() => set.Add(null));

      Assert.AreEqual(0, set.Count);
    }

    [Test]
    public void TestAddSameKey()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
      });

      Assert.AreEqual(1, set.Count);

      Assert.IsFalse(set.Add(new TypedString("IMAP4rev1")));
      Assert.AreEqual(1, set.Count);
    }

    [Test]
    public void TestAddToReadOnlyList()
    {
      var set = new TestTypedStringSet(true, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.IsTrue(set.IsReadOnly);

      Assert.Throws<NotSupportedException>(() => set.Add(new TypedString("STARTTLS")));
    }

    [Test]
    public void TestAddRange()
    {
      var set = new TestTypedStringSet(false, Enumerable.Empty<TypedString>());

      Assert.IsFalse(set.Contains("IMAP4rev1"));
      Assert.AreEqual(0, set.Count);

      set.AddRange(new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.AreEqual(3, set.Count);

      Assert.IsFalse(set.Add(new TypedString("IMAP4rev1")));
      Assert.IsFalse(set.Add(new TypedString("UIDPLUS")));
      Assert.IsFalse(set.Add(new TypedString("AUTH=PLAIN")));
    }

    [Test]
    public void TestAddRange_CollectionContainsNull()
    {
      var set = new TestTypedStringSet(false, Enumerable.Empty<TypedString>());

      Assert.IsFalse(set.Contains("IMAP4rev1"));
      Assert.AreEqual(0, set.Count);

      Assert.Throws<ArgumentException>(() => {
        set.AddRange(new[] {
          new TypedString("IMAP4rev1"),
          null,
          new TypedString("AUTH=PLAIN"),
        });
      });

      // XXX
      Assert.IsTrue(set.Contains(new TypedString("IMAP4rev1")));
      Assert.AreEqual(1, set.Count);
    }

    [Test]
    public void TestRemove()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.AreEqual(3, set.Count, "initial count");

      Assert.IsTrue(set.Remove(new TypedString("imap4rev1")), "Remove #0");
      Assert.AreEqual(2, set.Count, "Count #0");

      Assert.IsFalse(set.Remove(new TypedString("imap4rev1")), "Remove #1");
      Assert.AreEqual(2, set.Count, "Count #1");

      CollectionAssert.AreEquivalent(ToTypedStrings(new[] {"UIDPLUS", "AUTH=PLAIN"}), set, "equivalent Remove #1");
    }

    [Test]
    public void TestRemoveNull()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.Throws<ArgumentNullException>(() => set.Remove(null));

      Assert.AreEqual(3, set.Count);
    }

    [Test]
    public void TestRemoveWhere()
    {
      var set = new TestTypedStringSet(false, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.IsFalse(set.IsReadOnly);

      Assert.AreEqual(1, set.RemoveWhere(delegate(TypedString s) {
        return s.Value.StartsWith("AUTH=");
      }));

      Assert.AreEqual(2, set.Count);
      Assert.IsFalse(set.Contains("AUTH=PLAIN"));
    }

    [Test]
    public void TestRemoveWhereFromReadOnlyList()
    {
      var set = new TestTypedStringSet(true, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.IsTrue(set.IsReadOnly);

      Assert.Throws<NotSupportedException>(delegate {
        set.RemoveWhere(delegate(TypedString s) {
          return 0 < s.Value.Length;
        });
      });
    }

    [Test]
    public void TestRemoveFromReadOnlyList()
    {
      var set = new TestTypedStringSet(true, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.IsTrue(set.IsReadOnly);

      Assert.Throws<NotSupportedException>(() => set.Remove(new TypedString("IMAP4rev1")));
    }

    [Test]
    public void TestClearReadOnlyList()
    {
      var set = new TestTypedStringSet(true, new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      });

      Assert.IsTrue(set.IsReadOnly);

      Assert.Throws<NotSupportedException>(() => set.Clear());
    }

    [Test]
    public void TestGetEnumerator()
    {
      var values = new[] {
        new TypedString("IMAP4rev1"),
        new TypedString("UIDPLUS"),
        new TypedString("AUTH=PLAIN"),
      };

      var set = new TestTypedStringSet(false, values);
      var count = 0;

      foreach (var val in set) {
        count++;
      }

      Assert.AreEqual(3, count);
    }

    private IEnumerable<TypedString> ToTypedStrings(IEnumerable<string> enumerable)
    {
      return enumerable.Select(delegate(string s) {
        return new TypedString(s);
      });
    }

    [Test]
    public void TestExceptWith()
    {
      var setInitial = ToTypedStrings(new[] {"a", "c", "d", "b", "e"}).ToArray();
      var s = new TestTypedStringSet(false, setInitial);

      s.ExceptWith(new TypedString[0]);

      Assert.AreEqual(5, s.Count, "Count ExceptWith #0");
      CollectionAssert.AreEquivalent(setInitial, s, "equivalent ExceptWith #0");

      s = new TestTypedStringSet(false, setInitial);
      s.ExceptWith(ToTypedStrings(new[] {"b", "d", "e", "x", "y"}));

      Assert.AreEqual(2, s.Count, "Count ExceptWith #1");
      CollectionAssert.AreEquivalent(ToTypedStrings(new[] {"a", "c"}), s, "equivalent ExceptWith #1");
    }

    [Test]
    public void TestIntersectWith()
    {
      var setInitial = ToTypedStrings(new[] {"a", "c", "d", "b", "e"}).ToArray();
      var s = new TestTypedStringSet(false, setInitial);

      s.IntersectWith(new TypedString[0]);

      Assert.AreEqual(0, s.Count, "Count IntersectWith #0");
      CollectionAssert.AreEquivalent(new int[] {}, s, "equivalent IntersectWith #0");

      s = new TestTypedStringSet(false, setInitial);
      s.IntersectWith(ToTypedStrings(new[] {"b", "d", "e", "x", "y"}));

      Assert.AreEqual(3, s.Count, "Count IntersectWith #1");
      CollectionAssert.AreEquivalent(ToTypedStrings(new[] {"b", "d", "e"}).ToArray(), s, "equivalent IntersectWith #1");
    }

    [Test]
    public void TestSymmetricExceptWith()
    {
      var setInitial = ToTypedStrings(new[] {"a", "c", "d", "b", "e"}).ToArray();
      var s = new TestTypedStringSet(false, setInitial);

      s.SymmetricExceptWith(new TypedString[0]);

      Assert.AreEqual(5, s.Count, "Count SymmetricExceptWith #0");
      CollectionAssert.AreEquivalent(setInitial, s, "equivalent SymmetricExceptWith #0");

      s.SymmetricExceptWith(ToTypedStrings(new[] {"b", "d", "e", "x", "y"}));

      Assert.AreEqual(4, s.Count, "Count SymmetricExceptWith #1");
      CollectionAssert.AreEquivalent(ToTypedStrings(new[] {"a", "c", "x", "y"}).ToArray(), s, "equivalent SymmetricExceptWith #1");
    }

    [Test]
    public void TestUnionWith()
    {
      var setInitial = ToTypedStrings(new[] {"a", "c", "d", "b", "e"}).ToArray();
      var s = new TestTypedStringSet(false, setInitial);

      s.UnionWith(new TypedString[0]);

      Assert.AreEqual(5, s.Count, "Count UnionWith #0");
      CollectionAssert.AreEquivalent(setInitial, s, "equivalent UnionWith #0");

      s.UnionWith(ToTypedStrings(new[] {"b", "d", "e", "x", "y"}));

      Assert.AreEqual(7, s.Count, "Count UnionWith #1");
      CollectionAssert.AreEquivalent(ToTypedStrings(new[] {"a", "b", "c", "d", "e", "x", "y"}).ToArray(), s, "equivalent UnionWith #1");
    }

    [Test]
    public void TestIsSubsetOf()
    {
      var s = new TestTypedStringSet(false, ToTypedStrings(new[] {"a", "c", "b"}));

      Assert.IsTrue(s.IsSubsetOf(ToTypedStrings(new[] {"A", "b", "C", "d"})), "IsSubsetOf #0");
      Assert.IsTrue(s.IsSubsetOf(ToTypedStrings(new[] {"A", "b", "C"})), "IsSubsetOf #1");
      Assert.IsFalse(s.IsSubsetOf(ToTypedStrings(new[] {"A", "b", "d", "x"})), "IsSubsetOf #2");
      Assert.IsFalse(s.IsSubsetOf(ToTypedStrings(new[] {"A", "b", "d"})), "IsSubsetOf #3");
    }

    [Test]
    public void TestIsSubsetOfEmpty()
    {
      var s = new TestTypedStringSet();

      Assert.IsTrue(s.IsSubsetOf(ToTypedStrings(new[] {"x"})), "IsSubsetOf #0");
      Assert.IsTrue(s.IsSubsetOf(ToTypedStrings(new string[0])), "IsSubsetOf #1");

      s = new TestTypedStringSet(false, ToTypedStrings(new[] {"x"}));

      Assert.IsFalse(s.IsSubsetOf(ToTypedStrings(new string[0])), "IsSubsetOf #2");
    }

    [Test]
    public void TestIsProperSubsetOf()
    {
      var s = new TestTypedStringSet(false, ToTypedStrings(new[] {"a", "c", "b"}));

      Assert.IsTrue(s.IsProperSubsetOf(ToTypedStrings(new[] {"A", "b", "C", "d"})), "IsProperSubsetOf #0");
      Assert.IsFalse(s.IsProperSubsetOf(ToTypedStrings(new[] {"A", "b", "C"})), "IsProperSubsetOf #1");
      Assert.IsFalse(s.IsProperSubsetOf(ToTypedStrings(new[] {"A", "b", "d", "x"})), "IsProperSubsetOf #2");
      Assert.IsFalse(s.IsProperSubsetOf(ToTypedStrings(new[] {"A", "b", "d"})), "IsProperSubsetOf #3");
    }

    [Test]
    public void TestIsProperSubsetOfEmpty()
    {
      var s = new TestTypedStringSet();

      Assert.IsFalse(s.IsProperSubsetOf(ToTypedStrings(new[] {"x"})), "IsProperSubsetOf #0");
      Assert.IsTrue(s.IsProperSubsetOf(ToTypedStrings(new string[0])), "IsProperSubsetOf #1");

      s = new TestTypedStringSet(false, ToTypedStrings(new[] {"x"}));

      Assert.IsFalse(s.IsProperSubsetOf(ToTypedStrings(new string[0])), "IsProperSubsetOf #2");
    }

    [Test]
    public void TestIsSupersetOf()
    {
      var s = new TestTypedStringSet(false, ToTypedStrings(new[] {"a", "c", "b"}));

      Assert.IsTrue(s.IsSupersetOf(ToTypedStrings(new[] {"A", "b"})), "IsSupersetOf #0");
      Assert.IsTrue(s.IsSupersetOf(ToTypedStrings(new[] {"A", "b", "C"})), "IsSupersetOf #1");
      Assert.IsFalse(s.IsSupersetOf(ToTypedStrings(new[] {"A", "b", "d", "x"})), "IsSupersetOf #2");
      Assert.IsFalse(s.IsSupersetOf(ToTypedStrings(new[] {"A", "b", "d"})), "IsSupersetOf #3");
    }

    [Test]
    public void TestIsSupersetOfEmpty()
    {
      var s = new TestTypedStringSet();

      Assert.IsFalse(s.IsSupersetOf(ToTypedStrings(new[] {"x"})), "IsSupersetOf #0");
      Assert.IsTrue(s.IsSupersetOf(ToTypedStrings(new string[0])), "IsSupersetOf #1");

      s = new TestTypedStringSet(false, ToTypedStrings(new[] {"x"}));

      Assert.IsTrue(s.IsSupersetOf(ToTypedStrings(new string[0])), "IsSupersetOf #2");
    }

    [Test]
    public void TestIsProperSupersetOf()
    {
      var s = new TestTypedStringSet(false, ToTypedStrings(new[] {"a", "c", "b"}));

      Assert.IsTrue(s.IsProperSupersetOf(ToTypedStrings(new[] {"A", "b"})), "IsProperSupersetOf #0");
      Assert.IsFalse(s.IsProperSupersetOf(ToTypedStrings(new[] {"A", "b", "C"})), "IsProperSupersetOf #1");
      Assert.IsFalse(s.IsProperSupersetOf(ToTypedStrings(new[] {"A", "b", "d", "x"})), "IsProperSupersetOf #2");
      Assert.IsFalse(s.IsProperSupersetOf(ToTypedStrings(new[] {"A", "b", "d"})), "IsProperSupersetOf #3");
    }

    [Test]
    public void TestIsProperSupersetOfEmpty()
    {
      var s = new TestTypedStringSet();

      Assert.IsFalse(s.IsProperSupersetOf(ToTypedStrings(new[] {"x"})), "IsProperSupersetOf #0");
      Assert.IsFalse(s.IsProperSupersetOf(ToTypedStrings(new string[0])), "IsProperSupersetOf #1");

      s = new TestTypedStringSet(false, ToTypedStrings(new[] {"x"}));

      Assert.IsTrue(s.IsProperSupersetOf(ToTypedStrings(new string[0])), "IsProperSupersetOf #2");
    }

    [Test]
    public void TestOverlaps()
    {
      var s = new TestTypedStringSet(false, ToTypedStrings(new[] {"a", "c", "b"}));

      Assert.IsTrue(s.Overlaps(ToTypedStrings(new[] {"a", "B", "d"})), "Overlaps #0");
      Assert.IsFalse(s.Overlaps(ToTypedStrings(new[] {"D", "e"})), "Overlaps #1");
      Assert.IsFalse(s.Overlaps(ToTypedStrings(new string[0])), "Overlaps #2");
    }

    [Test]
    public void TestOverlapsEmpty()
    {
      var s = new TestTypedStringSet();

      Assert.IsFalse(s.Overlaps(ToTypedStrings(new[] {"x"})), "Overlaps #1");
      Assert.IsFalse(s.Overlaps(ToTypedStrings(new string[0])), "Overlaps #2");
    }

    [Test]
    public void TestSetEquals()
    {
      var s = new TestTypedStringSet(false, ToTypedStrings(new[] {"a", "b", "c"}));

      Assert.IsTrue(s.SetEquals(ToTypedStrings(new[] {"A", "b", "C"})), "SetEquals #0");
      Assert.IsFalse(s.SetEquals(ToTypedStrings(new[] {"A", "b", "D"})), "SetEquals #1");
      Assert.IsFalse(s.SetEquals(ToTypedStrings(new[] {"A", "b"})), "SetEquals #2");
      Assert.IsFalse(s.SetEquals(ToTypedStrings(new string[0])), "SetEquals #3");
    }
  }
}
