// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.DataStructures;
using Smdn.Net.Pop3.Protocol.Transmission;

namespace Smdn.Net.Pop3.Client {
  partial class PopClient {
    private const bool defaultGetUniqueId = false;

    public bool DeleteAfterRetrieve {
      get; set;
    }

    /// <exception cref="InvalidOperationException">throws if IsBusy is true</exception>
    public long MessageCount {
      get { ThrowIfNotConnected(); return EnsureDropListUpdated().MessageCount; }
    }

    /// <exception cref="InvalidOperationException">throws if IsBusy is true</exception>
    public long TotalSize {
      get { ThrowIfNotConnected(); return EnsureDropListUpdated().SizeInOctets; }
    }

    public void UpdateStatus()
    {
      RunExclusively(UpdateStatusCore);
    }

    private void UpdateStatusCore()
    {
      ThrowIfError(session.Stat(out var dropList));

      dropListing = dropList;
    }

    private PopDropListing EnsureDropListUpdated()
    {
      if (dropListing.HasValue)
        return dropListing.Value;

      RunExclusively(UpdateStatusCore);

      return dropListing.Value;
    }

    private long GetMessageCount()
    {
      if (!dropListing.HasValue)
        UpdateStatusCore();

      return dropListing.Value.MessageCount;
    }

    public IReadOnlyList<PopMessage> GetMessages(bool getUniqueId = defaultGetUniqueId)
    {
      return RunExclusively(() => GetMessagesCore(getUniqueId));
    }

    // TODO: CancellationToken
    private IReadOnlyList<PopMessage> GetMessagesCore(bool getUniqueId)
    {
      ThrowIfError(session.ListAll(out var scanLists, (int)(dropListing?.MessageCount ?? 0L)));

      if (scanLists.Count == 0)
#if NET45
        return ArrayExtensions.Empty<PopMessage>();
#else
        return Array.Empty<PopMessage>();
#endif

      var ret = new List<PopMessage>(scanLists.Count);

      if (getUniqueId) {
        ThrowIfError(session.UidlAll(out var uids, scanLists.Count));

        var uidLists = new Dictionary<long, PopUniqueIdListing>(scanLists.Count);

        foreach (var uid in uids) {
          uidLists[uid.MessageNumber] = uid;
        }

        foreach (var scanList in scanLists) {
          if (uidLists.TryGetValue(scanList.MessageNumber, out var uniqueIdList)) {
            ret.Add(ToMessage(scanList, uniqueIdList));
          }
          else {
            var message = ToMessage(scanList, null);

            message.UniqueIdList = null; // XXX: set isUniqueIdListed true

            ret.Add(message);
          }
        }
      }
      else {
        foreach (var scanList in scanLists) {
          ret.Add(ToMessage(scanList, null));
        }
      }

      return ret;
    }

    public PopMessage GetMessage(long messageNumber, bool getUniqueId = defaultGetUniqueId)
    {
      if (messageNumber <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(messageNumber), messageNumber);

      return RunExclusively(delegate {
        return GetMessageByMessageNumber(messageNumber, getUniqueId);
      });
    }

    // TODO: CancellationToken
    private PopMessage GetMessageByMessageNumber(long messageNumber, bool getUniqueId)
    {
      /*
       * find exist
       */
      var message = FindMessageByMessageNumber(messageNumber);

      if (message != null) {
        message.ThrowIfMarkedAsDeleted();

        if (!getUniqueId || message.UniqueIdList.HasValue)
          return message;
      }

      /*
       * list newly
       */
      if (GetMessageCount() < messageNumber)
        throw new PopMessageNotFoundException(messageNumber);

      PopScanListing scanList;

      var result = session.List(messageNumber, out scanList);

      if (result.IsResponseStatus(PopStatusIndicator.Negative))
        throw new PopMessageNotFoundException(messageNumber);
      else
        ThrowIfError(result);

      if (getUniqueId) {
        ThrowIfError(session.Uidl(messageNumber, out var uniqueIdList));

        return ToMessage(scanList, uniqueIdList);
      }
      else {
        return ToMessage(scanList, null);
      }
    }

    public PopMessage GetFirstMessage(bool getUniqueId = defaultGetUniqueId)
    {
      return RunExclusively(delegate {
        return GetMessageByMessageNumber(1L, getUniqueId);
      });
    }

    public PopMessage GetLastMessage(bool getUniqueId = defaultGetUniqueId)
    {
      return RunExclusively(delegate {
        return GetLastMessageCore(getUniqueId);
      });
    }

    private PopMessage GetLastMessageCore(bool getUniqueId)
    {
      var lastMessageNumber = GetMessageCount();

      if (lastMessageNumber == 0L)
        throw new PopMessageNotFoundException(1L);
      else
        return GetMessageByMessageNumber(lastMessageNumber, getUniqueId);
    }

    public PopMessage GetMessage(string uniqueId)
    {
      uniqueId = GetValidatedUniqueId(uniqueId);

      return RunExclusively(delegate {
        return GetMessageByUniqueId(uniqueId);
      });
    }

    // TODO: CancellationToken
    private PopMessage GetMessageByUniqueId(string validatedUniqueId)
    {
      /*
       * find exist
       */
      var message = FindMessageByUniqueId(validatedUniqueId);

      if (message != null) {
        message.ThrowIfMarkedAsDeleted();

        return message;
      }

      /*
       * list newly
       */
      ThrowIfError(session.UidlAll(out var uniqueIdLists));

      foreach (var uniqueIdList in uniqueIdLists) {
        if (!uniqueIdList.Equals(validatedUniqueId))
          continue;

        ThrowIfError(session.List(uniqueIdList.MessageNumber, out var scanList));

        return ToMessage(scanList, uniqueIdList);
      }

      throw new PopMessageNotFoundException(validatedUniqueId);
    }

    private PopMessage FindMessageByMessageNumber(long messageNumber)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      if (messages.TryGetValue(messageNumber, out var message))
        return message;
      else
        return null;
    }

    private PopMessage FindMessageByUniqueId(string uniqueId)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      foreach (var message in messages.Values) {
        if (message.UniqueIdList.HasValue && message.UniqueIdList.Value.Equals(uniqueId))
          return message;
      }

      return null;
    }

    private PopMessage ToMessage(PopScanListing scanList, PopUniqueIdListing? uniqueIdList)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      PopMessage message;

      if (messages.TryGetValue(scanList.MessageNumber, out message)) {
        // update values
        message.ScanList = scanList;
        message.UniqueIdList = uniqueIdList ?? message.UniqueIdList;
      }
      else {
        message = new PopMessage(this, scanList, uniqueIdList);

        messages[message.MessageNumber] = message;
      }

      return message;
    }

    public void CancelDelete()
    {
      RunExclusively(CancelDeleteCore);
    }

    private void CancelDeleteCore()
    {
      ThrowIfError(session.Rset());

      dropListing = null; // must be updated

      foreach (var message in messages.Values) {
        message.IsMarkedAsDeleted = false;
      }
    }

    public void KeepAlive()
    {
      RunExclusively(KeepAliveCore);
    }

    private void KeepAliveCore()
    {
      ThrowIfError(session.NoOp());
    }

    internal void Delete(PopMessage message)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      ThrowIfError(session.Dele(message.MessageNumber));

      dropListing = null; // must be updated
    }

    /*
     * utilities
     */
    internal static string GetValidatedUniqueId(string uniqueId)
    {
      if (uniqueId == null)
        throw new ArgumentNullException(nameof(uniqueId));
      if (uniqueId.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(uniqueId));

      return uniqueId;
    }

    internal static void ThrowIfError(PopCommandResult result)
    {
      if (!result.IsSuccess)
        throw new PopErrorResponseException(result);
    }

    private PopDropListing? dropListing = null;
    private Dictionary<long, PopMessage> messages = new Dictionary<long, PopMessage>();
  }
}
