// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Threading;

using Smdn.Formats;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Pop3.Protocol.Transmission.Transaction {
  /*
   * http://tools.ietf.org/html/rfc5034
   * RFC 5034 - The Post Office Protocol (POP3) Simple Authentication and Security Layer (SASL) Authentication Mechanism
   */
  internal sealed class AuthTransaction : PopTransactionBase, IPopExtension {
    PopCapability IPopExtension.RequiredCapability {
      get { return requiredCapability; }
    }

    public AuthTransaction(PopConnection connection,
                           SaslClientMechanism authMechanism,
                           CancellationToken cancellationToken)
      : base(connection)
    {
      this.authMechanism = authMechanism ?? throw new ArgumentNullException(nameof(authMechanism));
      this.requiredCapability = null; // do not check capability
      this.cancellationToken = cancellationToken;
    }

    /*
     * 4. The AUTH Command
     * 
     *    AUTH mechanism [initial-response]
     * 
     *       Arguments:
     * 
     *          mechanism: A string identifying a SASL authentication
     *          mechanism.
     * 
     *          initial-response: An optional initial client response, as
     *          defined in Section 3 of [RFC4422].  If present, this response
     *          MUST be encoded as Base64 (specified in Section 4 of
     *          [RFC4648]), or consist only of the single character "=", which
     *          represents an empty initial response.
     * 
     *       Restrictions:
     * 
     *          After an AUTH command has been successfully completed, no more
     *          AUTH commands may be issued in the same session.  After a
     *          successful AUTH command completes, a server MUST reject any
     *          further AUTH commands with an -ERR reply.
     * 
     *          The AUTH command may only be given during the AUTHORIZATION
     *          state.
     */
    internal class SaslInitialExchangeFailedException : InvalidOperationException {
      public SaslInitialExchangeFailedException(SaslClientMechanism causedSaslMechanism)
        : base($"SASL client did not generate initial-response (name: {causedSaslMechanism.Name}, status: {causedSaslMechanism.ExchangeStatus})")
      {
      }
    }

    protected override PopCommand PrepareCommand()
    {
      authMechanism.Initialize();

      byte[] initialClientResponse = null;

      if (authMechanism.ClientFirst) {
        if (authMechanism.GetInitialResponse(out initialClientResponse) == SaslExchangeStatus.Failed ||
            initialClientResponse == null)
          throw new SaslInitialExchangeFailedException(authMechanism);
      }

      if (initialClientResponse == null)
        return PopCommand.Create("AUTH",
                                 RequestArguments["mechanism"]);
      else
        /*
         * The optional initial-response argument to the AUTH command is
         * used to save a round trip when using authentication mechanisms
         * that support an initial client response.
         */
        return PopCommand.Create("AUTH",
                                 RequestArguments["mechanism"],
                                 Base64.GetEncodedString(initialClientResponse));
    }

    protected override void OnContinuationRequestReceived(PopContinuationRequest continuationRequest)
    {
      /*
       * A client response consists of a line containing a string
       * encoded as Base64.  If the client wishes to cancel the
       * authentication exchange, it issues a line with a single "*".
       * If the server receives such a response, it MUST reject the AUTH
       * command by sending an -ERR reply.
       * 
       * The optional initial-response argument to the AUTH command is
       * used to save a round trip when using authentication mechanisms
       * that support an initial client response.  If the initial
       * response argument is omitted and the chosen mechanism requires
       * an initial client response, the server MUST proceed by issuing
       * an empty challenge, as defined in Section 3 of [RFC4422].  In
       * POP3, an empty server challenge is defined as a line with only
       * a "+", followed by a single space.  It MUST NOT contain any
       * other data.
       */
      if (cancellationToken.IsCancellationRequested) {
        Connection.SendCommand(PopCommand.CreateCancelResponse());
        return;
      }

      var serverChallenge = Base64.Decode(continuationRequest.Base64Text.Segment.Array,
                                          continuationRequest.Base64Text.Segment.Offset,
                                          continuationRequest.Base64Text.Segment.Count);
      var status = authMechanism.Exchange(serverChallenge, out var clientResponse);

      if (status == SaslExchangeStatus.Failed || clientResponse == null)
        Connection.SendCommand(PopCommand.CreateCancelResponse());
      else
        Connection.SendCommand(PopCommand.CreateContinuation(Base64.GetEncodedString(clientResponse)));
    }

    private readonly SaslClientMechanism authMechanism;
    private readonly PopCapability requiredCapability;
    private readonly CancellationToken cancellationToken;
  }
}
