// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Diagnostics;
using System.IO;
using System.Net.Sockets;
using System.Threading;

using Smdn.Net.Connection;
using Smdn.IO.Streams.LineOriented;

namespace Smdn.Net.Pop3.Protocol.Transmission {
  public sealed class PopConnection : ConnectionBase {
    /*
     * class members
     */
    public static TraceSource ConnectionTraceSource { get; } = new TraceSource("POP");

    /*
     * properties
     */
    protected override TraceSource TraceSource => ConnectionTraceSource;

    internal bool HandleResponseAsMultiline {
      get { return receiver.HandleAsMultiline; }
      set { receiver.HandleAsMultiline = value; }
    }

    protected override int GetDefaultPort(bool asSecurePortConnection)
      => asSecurePortConnection ? PopDefaultPorts.Pops : PopDefaultPorts.Pop;

    protected override void OnConnectionStreamChanged(Stream stream)
    {
      var bufferedStream = new LineOrientedBufferedStream(stream);

      receiver = new PopResponseReceiver(bufferedStream);
      sender = new PopCommandSender(bufferedStream);
    }

    protected override Exception CreateConnectException(string message, Exception baseException)
      => new PopConnectionException(message, baseException);

    protected override Exception CreateUpgradeToSecureConnectionException(string message, Exception baseException)
      => new PopSecureConnectionException(message, baseException);

    protected override Exception CreateUpgradeConnectionException(string message, Exception baseException)
      => new PopUpgradeConnectionException(message, baseException);

#region "seinding / receiving"
    public void SendCommand(PopCommand command)
    {
      ThrowIfNotConnected();

      try {
        sender.Send(command);
      }
      catch (SocketException ex) {
        throw CreateSocketErrorException(ex, false);
      }
      catch (IOException ex) {
        var socketException = ex.InnerException as SocketException;

        if (socketException == null)
          // unexpected IO error
          throw;
        else
          throw CreateSocketErrorException(socketException, false);
      }
    }

    public PopResponse TryReceiveResponse()
    {
      ThrowIfNotConnected();

      try {
        // receive and parse response
        return receiver.ReceiveResponse();
      }
      catch (SocketException ex) {
        throw CreateSocketErrorException(ex, true);
      }
      catch (IOException ex) {
        var socketException = ex.InnerException as SocketException;

        if (socketException == null)
          // unexpected IO error
          throw;
        else
          throw CreateSocketErrorException(socketException, true);
      }
    }

    public bool TryReceiveLine(Stream stream, out int receivedLineLength)
    {
      ThrowIfNotConnected();

      try {
        return receiver.ReceiveLine(stream, out receivedLineLength);
      }
      catch (IOException ex) {
        var socketException = ex.InnerException as SocketException;

        if (socketException == null)
          // unexpected IO error
          throw;
        else
          throw CreateSocketErrorException(socketException, true);
      }
    }

    private Exception CreateSocketErrorException(SocketException ex, bool receiving)
    {
      switch (ex.SocketErrorCode) {
        case SocketError.WouldBlock:
        case SocketError.TimedOut:
          // timeout error
          return new TimeoutException(receiving ? "receive timeout" : "send timeout", ex);

        default:
          // unexpected socket exception
          return new PopConnectionException(string.Format("unexpected socket error ({0})", ex.SocketErrorCode),
                                            ex);
      }
    }
#endregion

    private PopResponseReceiver receiver = null;
    private PopCommandSender sender = null;
  }
}
