using System;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Pop3.Protocol;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientOperationsTests {
    [Test]
    public void TestGetMessage()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        // get again
        Assert.AreSame(message, client.GetMessage(1L));
      });
    }

    [Test]
    public void TestGetFirstMessage()
    {
      GetFirstMessage(false);
    }

    [Test]
    public void TestGetFirstMessage_GetUniqueId()
    {
      GetFirstMessage(true);
    }

    private void GetFirstMessage(bool getUniqueId)
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        if (getUniqueId)
          server.EnqueueResponse("UIDL 1\r\n",
                                 "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        var message = getUniqueId
          ? client.GetFirstMessage(true)
          : client.GetFirstMessage();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(5L, client.MessageCount);
        Assert.AreEqual(1200L, client.TotalSize);

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        if (getUniqueId)
          Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // get again
        Assert.AreSame(message, getUniqueId ? client.GetFirstMessage(true) : client.GetFirstMessage());
      });
    }

    [Test]
    public void TestGetLastMessage()
    {
      GetLastMessage(false);
    }

    [Test]
    public void TestGetLastMessage_GetUniqueId()
    {
      GetLastMessage(true);
    }

    private void GetLastMessage(bool getUniqueId)
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 5\r\n",
                               "+OK 5 200\r\n");

        if (getUniqueId)
          server.EnqueueResponse("UIDL 5\r\n",
                                 "+OK 5 whqtswO00WBw418f9t5JxYwZ\r\n");

        var message = getUniqueId
          ? client.GetLastMessage(true)
          : client.GetLastMessage();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(5L, client.MessageCount);
        Assert.AreEqual(1200L, client.TotalSize);

        Assert.IsNotNull(message);
        Assert.AreEqual(5L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        if (getUniqueId)
          Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // get again
        Assert.AreSame(message, getUniqueId ? client.GetLastMessage(true) : client.GetLastMessage());
      });
    }

    [Test]
    [TestCase(false)]
    [TestCase(true)]
    public void TestGetLastMessage_StatFailed(bool getUniqueId)
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "-ERR\r\n");

        Assert.Throws<PopErrorResponseException>(() => client.GetLastMessage(getUniqueId));

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetFirstMessage_NoMessagesExist()
    {
      GetFirstOrLastMessage(delegate(PopClient client) {
        var ex = Assert.Throws<PopMessageNotFoundException>(() => client.GetFirstMessage());
        Assert.AreEqual(1L, ex.MessageNumber);
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetLastMessage_NoMessagesExist()
    {
      GetFirstOrLastMessage(delegate(PopClient client) {
        var ex = Assert.Throws<PopMessageNotFoundException>(() => client.GetLastMessage());
        Assert.AreEqual(1L, ex.MessageNumber);
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    private void GetFirstOrLastMessage(Action<PopClient> action)
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        action(client);
      });
    }

    [Test]
    public void TestGetMessageGetUniqueId()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("UIDL 1\r\n",
                               "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        var message = client.GetMessage(1L, true);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);
      });
    }

    [Test]
    public void TestGetMessageDeleted()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        message.MarkAsDeleted();

        Assert.IsTrue(message.IsMarkedAsDeleted);

        var ex = Assert.Throws<PopMessageDeletedException>(() => client.GetMessage(1L));

        Assert.AreEqual(1L, ex.MessageNumber);
      });
    }

    [Test]
    public void TestGetMessageNotFound()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 120\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "-ERR no such message\r\n");

        var ex = Assert.Throws<PopMessageNotFoundException>(() => client.GetMessage(1L));

        Assert.AreEqual(1L, ex.MessageNumber);
        Assert.IsNull(ex.UniqueId);

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageZero()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentOutOfRangeException>(() => client.GetMessage(0L));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageGreaterThanMessageCount()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 3 1024\r\n");

        var ex = Assert.Throws<PopMessageNotFoundException>(() => client.GetMessage(4L));
        Assert.AreEqual(4L, ex.MessageNumber);
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageNotConnected()
    {
      using (var client = new PopClient()) {
        Assert.Throws<InvalidOperationException>(() => client.GetMessage(1L));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      }
    }

    [Test]
    public void TestGetMessageReconnect()
    {
      var sessionMessages = new PopMessage[2];

      for (var i = 0; i < 2; i++) {
        TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
          server.EnqueueResponse("STAT\r\n",
                                 "+OK 1 200\r\n");
          server.EnqueueResponse("LIST 1\r\n",
                                 "+OK 1 200\r\n");

          var message = client.GetMessage(1L);

          Assert.IsFalse(client.IsBusy, "IsBusy");

          Assert.AreEqual(1L, client.MessageCount);
          Assert.AreEqual(200L, client.TotalSize);

          Assert.IsNotNull(message);
          Assert.AreEqual(1L, message.MessageNumber);
          Assert.AreEqual(200L, message.Length);

          sessionMessages[i] = message;
        });
      }

      Assert.AreNotEqual(sessionMessages[0], sessionMessages[1]);
    }

    [Test]
    public void TestGetMessageByUid()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");
        server.EnqueueResponse("LIST 2\r\n",
                               "+OK 2 240\r\n");

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(2L, message.MessageNumber);
        Assert.AreEqual(240L, message.Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", message.UniqueId);
      });
    }

    [Test]
    public void TestGetMessageByUidError()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("UIDL\r\n",
                               "-ERR\r\n");

        Assert.Throws<PopErrorResponseException>(() => client.GetMessage("QhdPYR:00WBw1Ph7x7"));

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageByUidDeleted()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");
        server.EnqueueResponse("LIST 2\r\n",
                               "+OK 2 240\r\n");

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        server.EnqueueResponse("DELE 2\r\n",
                               "+OK\r\n");

        message.MarkAsDeleted();

        var ex = Assert.Throws<PopMessageDeletedException>(() => client.GetMessage(message.UniqueId));

        Assert.AreEqual(message.MessageNumber, ex.MessageNumber);

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageByUidAlreadyListed()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true);

        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.IsNotNull(message);
        Assert.AreEqual(2L, message.MessageNumber);
        Assert.AreSame(messages[1], message);
      });
    }

    [Test]
    public void TestGetMessageByUidNull()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentNullException>(() => client.GetMessage((string)null));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageByUidEmpty()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentException>(() => client.GetMessage(string.Empty));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageByUidNotFound()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var ex = Assert.Throws<PopMessageNotFoundException>(() => client.GetMessage("non-existent"));

        Assert.AreEqual(0L, ex.MessageNumber);
        Assert.IsNotNull(ex.UniqueId);
        Assert.AreEqual("non-existent", ex.UniqueId);

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessages()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");

        var messages = client.GetMessages();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(3, messages.Count);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);

        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");

        var messagesSecond = client.GetMessages();

        Assert.AreSame(messages[0], messagesSecond[0]);
        Assert.AreSame(messages[1], messagesSecond[1]);
        Assert.AreSame(messages[2], messagesSecond[2]);
      });
    }

    [Test]
    public void TestGetMessagesGetUniqueId()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(3, messages.Count);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", messages[0].UniqueId);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);
        Assert.AreEqual("xxxxxxxxx", messages[2].UniqueId);
      });
    }

    [Test]
    public void TestGetMessagesEmpty()
    {
      GetMessagesEmpty(false);
    }

    [Test]
    public void TestGetMessagesEmptyGetUniqueId()
    {
      GetMessagesEmpty(true);
    }

    private void GetMessagesEmpty(bool getUniqueId)
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               ".\r\n");

        var messages = getUniqueId ? client.GetMessages(true) : client.GetMessages();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(0, messages.Count);
      });
    }

    [Test]
    public void TestGetMessagesGetUniqueIdExtraUniqueId()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               "4 yyyyyyyyyyy\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(3, messages.Count);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", messages[0].UniqueId);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);
        Assert.AreEqual("xxxxxxxxx", messages[2].UniqueId);
      });
    }

    [Test]
    public void TestGetMessagesGetUniqueIdMissingUniqueId()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(3, messages.Count);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", messages[0].UniqueId);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);
        Assert.IsNull(messages[1].UniqueId);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);
        Assert.AreEqual("xxxxxxxxx", messages[2].UniqueId);
      });
    }

    [Test]
    public void TestUpdateStatus()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        client.UpdateStatus();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        // re-issue
        server.EnqueueResponse("STAT\r\n",
                               "+OK 2 400\r\n");

        client.UpdateStatus();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(2L, client.MessageCount);
        Assert.AreEqual(400L, client.TotalSize);
      });
    }

    [Test]
    public void TestUpdateStatus_LoggedOut()
    {
      TestCases.WithLoggedOutClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.UpdateStatus());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestUpdateStatus_Disconnected()
    {
      TestCases.WithDisconnectedClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<Smdn.Net.Pop3.Protocol.PopConnectionException>(() => client.UpdateStatus());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestCancelDelete()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        message.MarkAsDeleted();

        Assert.IsTrue(message.IsMarkedAsDeleted);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        Assert.AreEqual(0L, client.MessageCount);
        Assert.AreEqual(0L, client.TotalSize);

        server.EnqueueResponse("RSET\r\n",
                               "+OK\r\n");

        client.CancelDelete();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.IsFalse(message.IsMarkedAsDeleted);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);
      });
    }

    [Test]
    public void TestCancelDelete_MessageNotListed()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("RSET\r\n",
                               "+OK\r\n");

        client.CancelDelete();

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestCancelDelete_LoggedOut()
    {
      TestCases.WithLoggedOutClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.CancelDelete());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestCancelDelete_Disconnected()
    {
      TestCases.WithDisconnectedClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<Smdn.Net.Pop3.Protocol.PopConnectionException>(() => client.CancelDelete());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestKeepAlive()
    {
      TestCases.WithAuthenticatedClient(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("NOOP\r\n",
                               "+OK\r\n");

        client.KeepAlive();

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestKeepAlive_LoggedOut()
    {
      TestCases.WithLoggedOutClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.KeepAlive());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestKeepAlive_Disconnected()
    {
      TestCases.WithDisconnectedClient(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<Smdn.Net.Pop3.Protocol.PopConnectionException>(() => client.KeepAlive());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestInternalImpl_RunExclusively()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        TestUtils.Repeat(10, () => {
          var responded = 0;
          var responseTask = Task.Run(() => {
            server.WaitForClientRequest();

            Thread.Sleep(100);

            server.EnqueueResponse("NOOP\r\n",
                                   "+OK\r\n",
                                   () => Interlocked.Exchange(ref responded, 1),
                                   null);
          });

          var t = Task.Run(() => client.KeepAlive()); // call RunExclusively()

          server.WaitForClientRequest(); // wait for KeepAlive() started

          var count = 0;

          for (;;) {
            if (1 == Interlocked.CompareExchange(ref responded, 1, 1))
              break;

            Assert.Throws<InvalidOperationException>(() => client.KeepAlive());

            if (1 == Interlocked.CompareExchange(ref responded, 1, 1))
              break;

            Assert.Throws<InvalidOperationException>(() => client.KeepAliveAsync());

            count++;

            Thread.Sleep(1);
          }

          Assert.DoesNotThrow(() => t.Wait());

          Assert.Greater(count, 0, "count of method call");

          responseTask.Wait();
        });
      });
    }

    [Test]
    public void TestInternalImpl_StartTaskExclusively()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        TestUtils.Repeat(10, () => {
          var responded = 0;
          var responseTask = Task.Run(() => {
            Thread.Sleep(100);

            server.EnqueueResponse("NOOP\r\n",
                                   "+OK\r\n",
                                   () => Interlocked.Exchange(ref responded, 1),
                                   null);
          });

          var t = client.KeepAliveAsync(); // call StartTaskExclusively()

          var count = 0;

          for (;;) {
            if (1 == Interlocked.CompareExchange(ref responded, 1, 1))
              break;

            Assert.Throws<InvalidOperationException>(() => client.KeepAlive());

            if (1 == Interlocked.CompareExchange(ref responded, 1, 1))
              break;

            Assert.Throws<InvalidOperationException>(() => client.KeepAliveAsync());

            count++;

            Thread.Sleep(1);
          }

          Assert.DoesNotThrow(() => t.Wait());

          Assert.Greater(count, 0, "count of method call");

          responseTask.Wait();
        });
      });
    }
  }
}
