using System;
using System.Net;
using NUnit.Framework;

using Smdn.Net.Pop3.Protocol.Session;

namespace Smdn.Net.Pop3.Protocol.Transmission {
  [TestFixture]
  public class PopCommandResultTests {
    [Test]
    public void TestSerializeBinary_StatusResponse_Ok()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK Capability list follows\r\n" +
                               "TOP\r\n" +
                               ".\r\n");

        var result = session.Capa();

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsFalse(result.IsConcludedByClientSide);
          Assert.IsTrue(result.IsSuccess);

          Assert.AreEqual("\"Capability list follows\"", deserialized.Description);

          Assert.IsNotNull(deserialized.StatusResponse);
          Assert.AreEqual(PopStatusIndicator.Positive, deserialized.StatusResponse.Status);

          Assert.AreEqual("Capability list follows", deserialized.StatusResponse.Text);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          Assert.AreEqual(3, deserialized.ReceivedResponses.Count);
          Assert.IsInstanceOf<PopStatusResponse>(deserialized.ReceivedResponses[0]);
          Assert.IsInstanceOf<PopFollowingResponse>(deserialized.ReceivedResponses[1]);
          Assert.IsInstanceOf<PopTerminationResponse>(deserialized.ReceivedResponses[2]);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_StatusResponse_Err()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("USER frated\r\n",
                               "-ERR sorry, no mailbox for frated here\r\n");

        var result = session.User(new NetworkCredential("frated", "_"));

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsFalse(result.IsConcludedByClientSide);
          Assert.IsFalse(result.IsSuccess);

          Assert.AreEqual("\"sorry, no mailbox for frated here\"", deserialized.Description);

          Assert.IsNotNull(deserialized.StatusResponse);
          Assert.AreEqual(PopStatusIndicator.Negative, deserialized.StatusResponse.Status);

          Assert.AreEqual("sorry, no mailbox for frated here", deserialized.StatusResponse.Text);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          Assert.AreEqual(1, deserialized.ReceivedResponses.Count);
          Assert.IsInstanceOf<PopStatusResponse>(deserialized.ReceivedResponses[0]);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_ConcludedByClientSide_Success()
    {
      TestCases.WithLoggedInSession((session, server) => {
        var result = session.Login("user", "pass"); // already authenticated

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsTrue(result.IsConcludedByClientSide);
          Assert.IsTrue(result.IsSuccess);

          StringAssert.StartsWith("<", deserialized.Description);
          StringAssert.EndsWith(">", deserialized.Description);

          Assert.IsNull(deserialized.StatusResponse);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          CollectionAssert.IsEmpty(deserialized.ReceivedResponses);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_ConcludedByClientSide_Failure()
    {
      TestCases.WithConnectedSession(session => {
        var result = session.Login(new NullCredential());

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsTrue(result.IsConcludedByClientSide);
          Assert.IsFalse(result.IsSuccess);

          StringAssert.StartsWith("<", deserialized.Description);
          StringAssert.EndsWith(">", deserialized.Description);

          Assert.IsNull(deserialized.StatusResponse);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          CollectionAssert.IsEmpty(deserialized.ReceivedResponses);
        });
      });
    }
  }
}

