using System;
using NUnit.Framework;

namespace Smdn.Net.Pop3.Protocol {
  [TestFixture]
  public class PopCapabilityTests {
    [Test]
    public void TestEquals()
    {
      var x = PopCapability.Uidl;

      Assert.IsTrue(x.Equals(PopCapability.Uidl));
      Assert.IsTrue(x.Equals(new PopCapability("uidl")));
      Assert.IsTrue(x.Equals((object)"UIDL"));
      Assert.IsTrue(x.Equals((object)"uidl"));
      Assert.IsTrue(x.Equals((object)PopCapability.Uidl));
      Assert.IsTrue(x.Equals("UIDL"));
      Assert.IsTrue(x.Equals("uidl"));
      Assert.IsTrue(x.Equals(PopCapability.Uidl.Value));

      Assert.IsFalse(x.Equals(PopCapability.Top));
      Assert.IsFalse(x.Equals(new PopAuthenticationMechanism("UIDL")));
      Assert.IsFalse(x.Equals((object)new PopAuthenticationMechanism("UIDL")));
    }

    [Test]
    public void TestEquals_WithArguments()
    {
      var x = new PopCapability("X-FOO", "X-ARG1", "X-ARG2");

      Assert.IsTrue(x.Equals(new PopCapability("X-FOO", "X-ARG1", "X-ARG2")));
      Assert.IsTrue(x.Equals(new PopCapability("x-foo", "x-arg1", "x-arg2")));
      Assert.IsTrue(x.Equals("X-FOO X-ARG1 X-ARG2"));
      Assert.IsTrue(x.Equals("x-foo x-arg1 x-arg2"));

      Assert.IsFalse(x.Equals(new PopCapability("X-FOO", "X-ARG1")));
      Assert.IsFalse(x.Equals("X-FOO X-ARG1"));

      Assert.IsFalse(x.Equals(new PopCapability("X-FOO")));
      Assert.IsFalse(x.Equals("X-FOO"));
    }

    [Test]
    public void TestOpEquality()
    {
      var x = PopCapability.Uidl;

      Assert.IsTrue(x == PopCapability.Uidl);
      Assert.IsTrue(x == new PopCapability("uidl"));
      Assert.IsFalse(x != PopCapability.Uidl);
      Assert.IsFalse(x != new PopCapability("uidl"));

      Assert.IsFalse(x == PopCapability.Top);
      Assert.IsFalse(x == new PopCapability("top"));
      Assert.IsTrue(x != PopCapability.Top);
      Assert.IsTrue(x != new PopCapability("top"));
    }

    [Test]
    public void TestSet_TryGetValue()
    {
      Assert.IsTrue(PopCapability.AllCapabilities.TryGetValue("UIDL", out var capa1));
      Assert.AreSame(PopCapability.Uidl, capa1);

      Assert.IsTrue(PopCapability.AllCapabilities.TryGetValue("uidl", out var capa2));
      Assert.AreSame(PopCapability.Uidl, capa2);

      Assert.AreSame(capa1, capa2);
    }

    [Test]
    public void TestContainsAllArguments()
    {
      var capa = new PopCapability("sasl", "cram-md5", "plain");

      Assert.IsTrue(capa.ContainsAllArguments());
      Assert.IsTrue(capa.ContainsAllArguments("cram-md5"));
      Assert.IsTrue(capa.ContainsAllArguments("plain"));
      Assert.IsTrue(capa.ContainsAllArguments("cram-md5", "plain"));
      Assert.IsTrue(capa.ContainsAllArguments("plain", "cram-md5"));
      Assert.IsTrue(capa.ContainsAllArguments("PLAIN", "CRAM-MD5"));
      Assert.IsFalse(capa.ContainsAllArguments("plain", "cram-md5", "digest-md5"));
      Assert.IsFalse(capa.ContainsAllArguments("digest-md5"));
    }

    [Test]
    public void TestConstructor()
    {
      Assert.Throws<ArgumentNullException>(() => new PopCapability(null));
      Assert.Throws<ArgumentException>(() => new PopCapability(string.Empty));

      Assert.Throws<ArgumentException>(() => new PopCapability("tag", "arg1", null));
      Assert.Throws<ArgumentException>(() => new PopCapability("tag", "arg1", string.Empty));
    }

    [Test]
    public void TestValue()
    {
      var capa = new PopCapability("TOP");

      Assert.AreEqual("TOP", capa.Value);
    }

    [Test]
    public void TestToString()
    {
      Assert.AreEqual("UIDL", PopCapability.Uidl.ToString());
      Assert.AreEqual("uidl", new PopCapability("uidl").ToString());
      Assert.AreEqual("TOP", PopCapability.Top.ToString());
    }

    [Test]
    public void TestSerializeBinary()
    {
      var capa = new PopCapability("sasl", "cram-md5", "plain");

      Assert.IsTrue(capa.ContainsAllArguments());
      Assert.IsTrue(capa.ContainsAllArguments("cram-md5"));
      Assert.IsTrue(capa.ContainsAllArguments("plain"));
      Assert.IsTrue(capa.ContainsAllArguments("cram-md5", "plain"));
      Assert.IsTrue(capa.ContainsAllArguments("plain", "cram-md5"));
      Assert.IsFalse(capa.ContainsAllArguments("digest-md5"));

      TestUtils.Assert.IsSerializableBinaryFormat(capa, delegate(PopCapability deserialized) {
        Assert.AreNotEqual(PopCapability.Sasl, deserialized);
        Assert.AreEqual(capa, deserialized);
        Assert.IsTrue(deserialized.ContainsAllArguments("cram-md5", "plain"));
      });
    }
  }
}
