using System;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Net.Pop3.Protocol {
  [TestFixture]
  public class PopCapabilitySetExtensionsTests {
    [Test]
    public void TestIsCapable_CapabilitiesNull()
    {
      PopPseudoCapabilitySet capabilities = null;

      Assert.Throws<ArgumentNullException>(() => capabilities.IsCapable(PopCapability.Sasl));
    }

    [Test]
    public void TestIsCapable_FeatureNull()
    {
      Assert.Throws<ArgumentNullException>(() => PopCapability.AllCapabilities.IsCapable((PopAuthenticationMechanism)null));
    }

    [Test]
    public void TestFindByTag()
    {
      var list = new PopPseudoCapabilitySet(
        PopCapability.Uidl,
        PopCapability.Pipelining,
        new PopCapability("SASL", "CRAM-MD5", "PLAIN")
      );

      Assert.AreEqual(PopCapability.Uidl, list.FindByTag(PopCapability.Uidl));
      Assert.AreEqual(PopCapability.Uidl, list.FindByTag("UIDL"));
      Assert.AreEqual(PopCapability.Uidl, list.FindByTag("uidl"));
      Assert.IsNotNull(list.FindByTag(PopCapability.Sasl));
      Assert.IsNull(list.FindByTag(PopCapability.Expire));
    }

    [Test]
    public void TestIsCapable()
    {
      var list = new PopPseudoCapabilitySet(
        PopCapability.Uidl,
        PopCapability.Pipelining,
        new PopCapability("SASL", "CRAM-MD5", "PLAIN")
      );

      Assert.IsTrue(list.IsCapable(PopCapability.Uidl));
      Assert.IsTrue(list.IsCapable(PopCapability.Sasl));
      Assert.IsTrue(list.IsCapable(new PopCapability("uidl")));
      Assert.IsTrue(list.IsCapable(new PopCapability("sasl")));
      Assert.IsTrue(list.IsCapable(new PopCapability("SASL", "CRAM-MD5")));
      Assert.IsTrue(list.IsCapable(new PopCapability("SASL", "PLAIN")));
      Assert.IsFalse(list.IsCapable(new PopCapability("SASL", "DIGEST-MD5")));
      Assert.IsFalse(list.IsCapable(PopCapability.Expire));
    }

    [Test]
    public void TestIsCapable_StandardFeatures()
    {
      Assert.IsTrue(PopPseudoCapabilitySet.Empty.IsCapable(PopResponseCode.InUse));
    }

    [Test]
    public void TestIsCapable_EnumValueWithExtension()
    {
      var feature = PopAuthenticationMechanism.CRAMMD5; // SASL CRAM-MD5

      Assert.IsFalse(PopPseudoCapabilitySet.Empty.IsCapable(feature));

      var capabilitiesListExtended = new PopPseudoCapabilitySet(new PopCapability("SASL", "CRAM-MD5", "DIGEST-MD5"));

      Assert.IsTrue(capabilitiesListExtended.IsCapable(feature));
    }
  }
}
