using System;
using System.Linq;
using System.Net;
using NUnit.Framework;

using Smdn.Text;

namespace Smdn.Security.Authentication.Sasl.Client {
  [TestFixture]
  public class SaslClientMechanismTests {
    [Test]
    public void TestGetAvailableMechanisms()
    {
      var mechanisms = SaslClientMechanism.GetAvailableMechanisms();

      Assert.IsNotNull(mechanisms);
      Assert.Greater(mechanisms.Count, 0);
      CollectionAssert.Contains(mechanisms, "LOGIN");
      CollectionAssert.Contains(mechanisms, "PLAIN");
      CollectionAssert.Contains(mechanisms, "ANONYMOUS");
      CollectionAssert.Contains(mechanisms, "DIGEST-MD5");
      CollectionAssert.Contains(mechanisms, "CRAM-MD5");
    }

    [Test]
    public void TestGetAvailableMechanisms_WithUsingMechanisms()
    {
      var mechanisms = SaslClientMechanism.GetAvailableMechanisms(new[] { "LOGIN", "login", "digest-md5", "X-UNKNOWN" });

      Assert.IsNotNull(mechanisms);
      Assert.AreEqual(2, mechanisms.Count);
      Assert.IsTrue(mechanisms.Contains("LOGIN", StringComparer.OrdinalIgnoreCase));
      Assert.IsTrue(mechanisms.Contains("DIGEST-MD5", StringComparer.OrdinalIgnoreCase));
    }

    [Test]
    public void TestGetAvailableMechanisms_WithUsingMechanisms_Empty()
    {
      var mechanisms = SaslClientMechanism.GetAvailableMechanisms(Enumerable.Empty<string>());

      Assert.IsNotNull(mechanisms);
      Assert.AreEqual(0, mechanisms.Count);
    }

    [Test]
    public void TestGetAvailableMechanisms_WithUsingMechanisms_Null()
    {
      var mechanisms = SaslClientMechanism.GetAvailableMechanisms(null);

      Assert.IsNotNull(mechanisms);
      Assert.AreEqual(0, mechanisms.Count);
    }

    [Test]
    public void TestCreate()
    {
      var mechanism = SaslClientMechanism.Create("PLAIN");

      Assert.IsNotNull(mechanism);
      Assert.IsInstanceOf<PlainMechanism>(mechanism);

      Assert.IsTrue(mechanism.IsPlainText);
      Assert.AreEqual(SaslExchangeStatus.None, mechanism.ExchangeStatus);
    }

    [Test]
    public void TestCreate_MechanismNameNull()
    {
      Assert.Throws<ArgumentNullException>(() => SaslClientMechanism.Create(null));
    }

    [Test]
    public void TestCreate_MechanismNameEmpty()
    {
      Assert.Throws<ArgumentException>(() => SaslClientMechanism.Create(string.Empty));
    }

    [Test]
    public void TestCreate_NotSupported()
    {
      var ex = Assert.Throws<SaslMechanismNotSupportedException>(() => SaslClientMechanism.Create("X-UNKNOWN"));

      Assert.AreEqual("X-UNKNOWN", ex.MechanismName);

      TestUtils.Assert.IsSerializableBinaryFormat(ex, deserialized => {
        Assert.AreEqual(ex.MechanismName, deserialized.MechanismName);
      });
    }

    [Test]
    public void TestIsMechanismPlainText()
    {
      Assert.IsTrue(SaslClientMechanism.IsMechanismPlainText("PLAIN"));
      Assert.IsTrue(SaslClientMechanism.IsMechanismPlainText("plain"));
      Assert.IsTrue(SaslClientMechanism.IsMechanismPlainText("LOGIN"));
      Assert.IsFalse(SaslClientMechanism.IsMechanismPlainText("CRAM-MD5"));
      Assert.IsFalse(SaslClientMechanism.IsMechanismPlainText("cram-md5"));
      Assert.IsFalse(SaslClientMechanism.IsMechanismPlainText("DIGEST-MD5"));
    }

    [Test]
    public void TestIsMechanismPlainText_MechanismNameNull()
    {
      Assert.Throws<ArgumentNullException>(() => SaslClientMechanism.IsMechanismPlainText(null));
    }

    [Test]
    public void TestIsMechanismPlainText_MechanismNameEmpty()
    {
      Assert.Throws<ArgumentException>(() => SaslClientMechanism.IsMechanismPlainText(string.Empty));
    }

    [Test]
    public void TestIsMechanismPlainText_NotSupported()
    {
      var ex = Assert.Throws<SaslMechanismNotSupportedException>(() => SaslClientMechanism.IsMechanismPlainText("X-UNKNOWN"));

      Assert.AreEqual("X-UNKNOWN", ex.MechanismName);

      TestUtils.Assert.IsSerializableBinaryFormat(ex, deserialized => {
        Assert.AreEqual(ex.MechanismName, deserialized.MechanismName);
      });
    }

    [SaslMechanism(MechanismName, true)]
    private class XThrowsInvalidOperationExceptionMechanism : SaslClientMechanism {
      public const string MechanismName = "X-THROWS-EXCEPTION-INVALID";

      public XThrowsInvalidOperationExceptionMechanism()
      {
        throw new InvalidOperationException();
      }

      protected override SaslExchangeStatus Exchange (ByteString serverChallenge, out ByteString clientResponse)
      {
        throw new NotImplementedException();
      }
    }

    [Test]
    public void TestCreate_TargetInvocationExceptionThrown()
    {
      SaslClientMechanism.RegisterMechanism(typeof(XThrowsInvalidOperationExceptionMechanism));

      var ex = Assert.Throws<SaslException>(() => SaslClientMechanism.Create(XThrowsInvalidOperationExceptionMechanism.MechanismName));

      Assert.IsNotNull(ex.InnerException);
      Assert.IsInstanceOf<InvalidOperationException>(ex.InnerException);
    }

    [SaslMechanism(MechanismName, true)]
    private class XThrowsNotSupportedExceptionMechanism : SaslClientMechanism {
      public const string MechanismName = "X-THROWS-EXCEPTION-NOTSUPPORTED";

      public XThrowsNotSupportedExceptionMechanism()
      {
        throw new NotSupportedException();
      }

      protected override SaslExchangeStatus Exchange(ByteString serverChallenge, out ByteString clientResponse)
      {
        throw new NotImplementedException();
      }
    }

    [Test]
    public void TestCreate_TargetInvocationExceptionThrown_NotSupportedException()
    {
      SaslClientMechanism.RegisterMechanism(typeof(XThrowsNotSupportedExceptionMechanism));

      var ex = Assert.Throws<SaslMechanismNotSupportedException>(() => SaslClientMechanism.Create(XThrowsNotSupportedExceptionMechanism.MechanismName));

      Assert.IsNotNull(ex.InnerException);
      Assert.AreEqual(XThrowsNotSupportedExceptionMechanism.MechanismName, ex.MechanismName);
      Assert.IsInstanceOf<NotSupportedException>(ex.InnerException);
    }

    [Test]
    public void TestGetCredentialAfterDisposed()
    {
      var mechanism = SaslClientMechanism.Create("PLAIN");

      mechanism.Credential = new NetworkCredential("test", "test");

      mechanism.Dispose();

      Assert.Throws<ObjectDisposedException>(() => Assert.IsNull(mechanism.Credential));
    }

    [Test]
    public void TestGetExchangeStatusAfterDisposed()
    {
      var mechanism = SaslClientMechanism.Create("PLAIN");

      mechanism.Credential = new NetworkCredential("test", "test");

      mechanism.Dispose();

      Assert.Throws<ObjectDisposedException>(() => Assert.AreEqual(SaslExchangeStatus.None, mechanism.ExchangeStatus));
    }

    [Test]
    public void TestExchangeAfterDisposed()
    {
      var mechanism = SaslClientMechanism.Create("PLAIN");

      mechanism.Credential = new NetworkCredential("test", "test");

      mechanism.Dispose();

      byte[] clientResponse;

      Assert.Throws<ObjectDisposedException>(() => mechanism.Exchange(null, out clientResponse));
    }

    [Test]
    public void TestRegisterNull()
    {
      Assert.Throws<ArgumentNullException>(() => SaslClientMechanism.RegisterMechanism(null));
    }

    [Test]
    public void TestRegisterUnassignableType()
    {
      Assert.Throws<ArgumentException>(() => SaslClientMechanism.RegisterMechanism(typeof(int)));
    }

    private abstract class AbstractCustomMechanism : SaslClientMechanism {
    }

    [Test]
    public void TestRegisterAbstractClass()
    {
      Assert.Throws<ArgumentException>(() => SaslClientMechanism.RegisterMechanism(typeof(AbstractCustomMechanism)));
    }

    private class UnnamedCustomMechanism : SaslClientMechanism {
      protected override SaslExchangeStatus Exchange (ByteString serverChallenge, out ByteString clientResponse)
      {
        throw new NotImplementedException();
      }
    }

    [Test]
    public void TestRegisterNameAttributeNotApplied()
    {
      Assert.Throws<ArgumentException>(() => SaslClientMechanism.RegisterMechanism(typeof(UnnamedCustomMechanism)));
    }

    [SaslMechanism("X-CUSTOM", false)]
    private class CustomMechanism : SaslClientMechanism {
      public CustomMechanism()
      {
      }

      protected override SaslExchangeStatus Exchange (ByteString serverChallenge, out ByteString clientResponse)
      {
        throw new NotImplementedException();
      }
    }

    [Test]
    public void TestRegister()
    {
      SaslClientMechanism.RegisterMechanism(typeof(CustomMechanism));

      CollectionAssert.Contains(SaslClientMechanism.GetAvailableMechanisms(), "X-CUSTOM");

      var inst = SaslClientMechanism.Create("X-CUSTOM");

      Assert.IsNotNull(inst);
      Assert.IsInstanceOf<CustomMechanism>(inst);
    }

    [SaslMechanism("LOGIN", true)]
    private class XLoginMechanism : SaslClientMechanism {
      public XLoginMechanism()
      {
      }

      protected override SaslExchangeStatus Exchange (ByteString serverChallenge, out ByteString clientResponse)
      {
        throw new NotImplementedException();
      }
    }

    [Test]
    public void TestRegisterReplace()
    {
      SaslClientMechanism inst;

      CollectionAssert.Contains(SaslClientMechanism.GetAvailableMechanisms(), "LOGIN");

      inst = SaslClientMechanism.Create("LOGIN");

      Assert.IsNotNull(inst);
      Assert.IsNotInstanceOf<XLoginMechanism>(inst);

      SaslClientMechanism.RegisterMechanism(typeof(XLoginMechanism));

      CollectionAssert.Contains(SaslClientMechanism.GetAvailableMechanisms(), "LOGIN");

      inst = SaslClientMechanism.Create("LOGIN");

      Assert.IsNotNull(inst);
      Assert.IsInstanceOf<XLoginMechanism>(inst);
    }
  }
}
